// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "assert/assert.hpp"
#include "event/Event.hpp"

namespace Dawn {
  /** Forwarders / Fake Virtuals for other methods. */
  template<typename...A>
  class StateEvent;
  class IStateEvent;
  class IStateOwner {
    public:
      virtual void _stateEventDisposed(IStateEvent *evt) = 0;
  };

  template<typename...A>
  struct StateEventListener {
    uint32_t id;
    IStateOwner *owner;
    std::function<void(A...)> listener;
    std::function<void(StateEventListener<A...>)> unsubWithParams;
    std::function<void()> unsub;
    StateEvent<A...> *event;
  };

  class IStateEvent {
    protected:
      /**
       * Received notification from a state owner to let this state event know
       * that it wishes to unsubscribe all the event listeners that it may have
       * attached to this event.
       * 
       * @param owner State owner that is being disposed.
       */
      virtual void _stateOwnerDestroyed(IStateOwner *owner) = 0;

      friend class StateOwner;
  };

  template<typename...A>
  class StateEvent : public IStateEvent {
    protected:
      uint32_t stateEventId = 0;

      /** Refer to IStateEvent::_stateOwnerDestroyed */
      void _stateOwnerDestroyed(IStateOwner *owner) override {
        auto it = this->_eventListeners.begin();
        while(it != this->_eventListeners.end()) {
          if(it->owner == owner) {
            it = this->_eventListeners.erase(it);
          } else {
            ++it;
          }
        }
      }

    public:
      std::vector<StateEventListener<A...>> _eventListeners;

      /**
       * Invokes the event and emits to all of the listeners.
       * 
       * @param args Arguments for this event to pass to the listeners.
       */
      void invoke(A... args) {
        auto it = this->_eventListeners.begin();
        while(it != this->_eventListeners.end()) {
          it->listener(args...);
          ++it;
        }
      }

      /**
       * Disposal of a state event.
       */
      ~StateEvent() {
        auto it = this->_eventListeners.begin();
        while(it != this->_eventListeners.end()) {
          it->owner->_stateEventDisposed(this);
          ++it;
        }
      }

      friend class StateOwner;
  };
}