// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "util/XmlParser.hpp"
#include "util/CodeGen.hpp"
#include "util/Language.hpp"

namespace Dawn {
  struct TextEventInfo {
    std::string character;
    std::string emotion;
    std::vector<struct LanguageString> strings;
    std::string key;
  };

  class TextStringParser : public XmlParser<struct LanguageString> {
    protected:
      std::vector<std::string> getRequiredAttributes() {
        return std::vector<std::string>{ "lang" };
      }

      std::map<std::string, std::string> getOptionalAttributes() {
        return std::map<std::string, std::string>();
      }

      int32_t onParse(
        Xml *node,
        std::map<std::string, std::string> values,
        struct LanguageString *out,
        std::string *error
      ) {
        out->lang = values["lang"];
        out->text = node->value;
        return 0;
      }
  };

  class TextEventParser : public XmlParser<struct TextEventInfo> {
    protected:
      std::vector<std::string> getRequiredAttributes() {
        return std::vector<std::string>{
          "character",
          "emotion"
        };
      }

      std::map<std::string, std::string> getOptionalAttributes() {
        return std::map<std::string, std::string>();
      }

      int32_t onParse(
        Xml *node,
        std::map<std::string, std::string> values,
        struct TextEventInfo *out,
        std::string *error
      ) {
        int32_t ret = 0;
        out->character = values["character"];
        out->emotion = values["emotion"];

        if(out->key.size() <= 0) {
          *error = "Text Event requries a language key to be defined.";
          return 1;
        }

        auto itChildren = node->children.begin();
        while(itChildren != node->children.end()) {
          auto c = *itChildren;
          if(c->node == "string") {
            struct LanguageString str;
            ret = (TextStringParser()).parse(c, &str, error);
            str.key = out->key;
            out->strings.push_back(str);
          }
          ++itChildren;
        }
        return ret;
      }
  };

  class TextEventGen : public CodeGen {
    public:
      static void generate(
        std::vector<std::string> *out,
        struct TextEventInfo *info,
        std::string tabs = ""
      ) {
        std::string emo = info->emotion;
        emo[0] = toupper(emo[0]);

        line(out, "new VisualNovelTextboxEvent(vnManager,", tabs);
        line(out, "this->" + info->character + "->vnCharacter, ", tabs + "  ");
        line(out, "this->" + info->character + "->emotion" + emo + ", ", tabs + "  ");
        line(out, "\"" + info->key  + "\"", tabs + "  ");
        line(out, ")", tabs);
      }
  };
}